from loguru import logger
from ..client.game_client import GameClient







class Gifts(GameClient):


    async def collect_citizen_gift(self, sync: bool = True) -> dict | bool:

        try:
            await self.send_json_message("irc", {})
            if sync:
                response = await self.wait_for_response("irc")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        


    async def collect_citizen_quest(self, choice: int, sync: bool = True) -> dict | bool:

        try:
            await self.send_json_message("jjc", {"CO": choice})
            if sync:
                response = await self.wait_for_response("jjc")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        


    async def collect_ressource_gift(self, resource_type: int, sync: bool = True) -> dict | bool:

        try:
            await self.send_json_message("rcc", {"RT": resource_type})
            if sync:
                response = await self.wait_for_response("rcc")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False


from ..client.game_client import GameClient
from loguru import logger







class Auth(GameClient):



    async def check_username_availability(
        self,
        name: str,
        sync: bool = True
    ) -> dict | bool:

        try:
            await self.send_json_message("vpn", {"PN": name})
            if sync:
                response = await self.wait_for_response("vpn")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False


    async def check_user_exists(
        self,
        name: str,
        sync: bool = True
    ) -> dict | bool:

        try:
            await self.send_json_message("vln", {"NOM": name})
            if sync:
                response = await self.wait_for_response("vln")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False
        

    async def register(
        self, 
        username: str, 
        email: str, 
        password: str,
        token: str, 
        sync: bool = True
    ) -> dict | bool:

        try:
            await self.send_json_message(
                "lre",
                {
                    "DID": 0,
                    "CONM": 175,
                    "RTM": 24,
                    "campainPId": -1,
                    "campainCr": -1,
                    "campainLP": -1,
                    "adID": -1,
                    "timeZone": 14,
                    "username": username,
                    "email": email,
                    "password": password,
                    "accountId": "1674256959939529708",
                    "ggsLanguageCode": "en",
                    "referrer": "https://empire.goodgamestudios.com",
                    "distributorId": 0,
                    "connectionTime": 175,
                    "roundTripTime": 24,
                    "campaignVars": ";https://empire.goodgamestudios.com;;;;;;-1;-1;;1674256959939529708;380635;;;;;",
                    "campaignVars_adid": "-1",
                    "campaignVars_lp": "-1",
                    "campaignVars_creative": "-1",
                    "campaignVars_partnerId": "-1",
                    "campaignVars_websiteId": "380635",
                    "timezone": 14,
                    "PN": username,
                    "PW": password,
                    "REF": "https://empire.goodgamestudios.com",
                    "LANG": "en",
                    "AID": "1674256959939529708",
                    "GCI": "",
                    "SID": 9,
                    "PLFID": 1,
                    "NID": 1,
                    "IC": "",
                    "RCT": token
                },
            )
            if sync:
                response = await self.wait_for_response("lre")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False
        

    async def login_with_token(
        self, 
        name: str, 
        password: str, 
        token: str, 
        sync: bool = True
    ) -> dict | bool:

        try:
            await self.send_json_message(
                "lli",
                {
                    "CONM": 175,
                    "RTM": 24,
                    "ID": 0,
                    "PL": 1,
                    "NOM": name,
                    "PW": password,
                    "LT": None,
                    "LANG": "fr",
                    "DID": "0",
                    "AID": "1674256959939529708",
                    "KID": "",
                    "REF": "https://empire.goodgamestudios.com",
                    "GCI": "",
                    "SID": 9,
                    "PLFID": 1,
                    "RCT": token,
                },
            )
            if sync:
                response = await self.wait_for_response("lli")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False


from ..client.game_client import GameClient
from loguru import logger







class Misc(GameClient):
    

    


    async def change_emblem(
        self,
        bg_type: int,
        bg_color_1: int,
        bg_color_2: int,
        icons_type: int,
        icon_id_1: int,
        icon_color_1: int,
        icon_id_2: int,
        icon_color_2: int,
        sync: bool = True
    ) -> dict | bool:

        try:
            await self.send_json_message(
                "cem",
                {
                    "CAE": {
                        "BGT": bg_type,
                        "BGC1": bg_color_1,
                        "BGC2": bg_color_2,
                        "SPT": icons_type,
                        "S1": icon_id_1,
                        "SC1": icon_color_1,
                        "S2": icon_id_2,
                        "SC2": icon_color_2,
                    }
                },
            )
            if sync:
                response = await self.wait_for_response("cem")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False  


from ..client.game_client import GameClient
from loguru import logger








class Tutorial(GameClient):
    
    
    
    async def choose_hero(
        self,
        hero_id: int = 802,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("hdc", {"HID": hero_id})
            if sync:
                response = await self.wait_for_response("hdc")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def skip_generals_intro(
        self,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("sgi", {})
            if sync:
                response = await self.wait_for_response("sgi")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def collect_noob_gift(
        self,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("uoa", {})
            if sync:
                response = await self.wait_for_response("uoa")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False


from ..client.game_client import GameClient
from loguru import logger









class Tax(GameClient):
    
    
    
    
    async def get_tax_infos(
        self,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("txi", {})
            if sync:
                response = await self.wait_for_response("txi")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def start_tax(
        self,
        tax_type: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "txs",
                {
                    "TT": tax_type,
                    "TX": 3
                }
            )
            if sync:
                response = await self.wait_for_response("txs")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        

    async def collect_tax(
        self,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("txc", {"TR": 29})
            if sync:
                response = await self.wait_for_response("txc")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False



from ..client.game_client import GameClient
from loguru import logger






class Soldiers(GameClient):
    
    
    
    async def get_recruitment_queue(self, sync: bool = True) -> dict | bool:
        
        try:
            
            await self.send_json_message("spl", {"LID":0})
            if sync:
                response = await self.wait_for_response("spl")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
     
        
    async def recruit_soldiers(
        self,
        castle_id: int,
        wod_id: int,
        amount: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            
            await self.send_json_message(
                "bup",
                {
                    "LID": 0,
                    "WID": wod_id,
                    "AMT": amount,
                    "PO": -1,
                    "PWR": 0,
                    "SK": 73,
                    "SID": 0,
                    "AID": castle_id
                }
            )
            if sync:
                response = await self.wait_for_response("bup")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False
        
        
        
        
    async def cancel_recruitment(
        self,
        slot_type: str,
        slot: int,
        sync: bool = True
        ) -> dict | bool:
        
        try:
            await self.send_json_message("mcu", {"LID":0, "S": slot, "ST": slot_type})
            if sync:
                response = await self.wait_for_response("mcu")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def recruitment_alliance_help(self, sync: bool = True) -> dict | bool:
        
        try:
            await self.send_json_message("ahr", {"ID":0, "T":6})
            if sync:
                response = await self.wait_for_response("ahr")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def get_units_inventory(
        self,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("gui", {})
            if sync:
                response = await self.wait_for_response("gui")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False        
        
        
    
    async def delete_units(
        self,
        wod_id: int,
        amount: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "dup",
                {
                    "WID": wod_id,
                    "A": amount,
                    "S": 0
                }
            )
            if sync:
                response = await self.wait_for_response("dup")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False   
        
        
        
    async def wait_receive_units(
        self,
        kingdom: int,
        castle_id: int,
        wod_id: int,
        amount: int,
        timeout: int = 5,
    ) -> dict | bool:

        try:
            response = await self.wait_for_response(
                "rue",
                {"AID": castle_id, "SID": kingdom, "WID": wod_id, "RUA": amount},
                timeout=timeout,
            )
            return response
        except Exception as e:
            logger.error(e)
            return False
 


from loguru import logger
import asyncio
import aiohttp






class Utils:
    
    API_IN_URL = "http://2captcha.com/in.php"
    API_RES_URL = "http://2captcha.com/res.php"
    
    
    
    
    def __init__(self):
        
        self.polling_interval = 5
        self.max_attempts = 20
        self.site_key = '6Lc7w34oAAAAAFKhfmln41m96VQm4MNqEdpCYm-k'
        self.site_url = 'https://empire.goodgamestudios.com/'
    
    

    
    
    
    async def get_recaptcha_token(self, api_key: str) -> str:
        
        payload = {
            'key': api_key,
            'method': 'userrecaptcha',
            'googlekey': self.site_key,
            'pageurl': self.site_url,
            'json': 1
        }

        async with aiohttp.ClientSession() as session:
            async with session.post(self.API_IN_URL, data=payload) as resp:
                data = await resp.json()
                if data.get('status') != 1:
                    logger.error(f"2Captcha submission error: {data}")
                    raise RuntimeError(f"Submit failed: {data.get('request')}")

                request_id = data.get('request')
                logger.info(f"Submitted CAPTCHA, request ID: {request_id}")

            for attempt in range(self.max_attempts):
                await asyncio.sleep(self.polling_interval)
                params = {
                    'key': api_key,
                    'action': 'get',
                    'id': request_id,
                    'json': 1
                }
                async with session.get(self.API_RES_URL, params=params) as res:
                    result = await res.json()
                    if result.get('status') == 1:
                        token = result.get('request')
                        logger.info("Retrieved token from 2Captcha")
                        return token
                    elif result.get('request') == 'CAPCHA_NOT_READY':
                        logger.debug(f"Attempt {attempt+1}/{self.max_attempts}: CAPTCHA not ready")
                        continue
                    else:
                        logger.error(f"2Captcha retrieval error: {result}")
                        raise RuntimeError(f"Get failed: {result.get('request')}")

        raise TimeoutError("2Captcha solving timed out")
    
    
    
    def skip_calculator(
        
        self,
        time: int,
        skip_type: list[str] = None
        
        ) -> list[str]:
        
        
        all_skip_values = [(86400, "MS7"), (18000, "MS6"), (3600, "MS5"), (1800, "MS4"), (600, "MS3"), (300, "MS2"), (60, "MS1")]
        
        if skip_type:
            skip_values = [(sec, lbl) for sec, lbl in all_skip_values if lbl in skip_type]
            if not skip_values:
                logger.error("Your skips are not allowed!")
                
        else:
            skip_values = all_skip_values
            
        minutes = time // 60
        skip_minutes = [(sec // 60, label) for sec, label in all_skip_values]
        INF = float('inf')
        dp = [INF] * (minutes + 1)
        prev = [None] * (minutes + 1)
        dp[0] = 0
        
        for i in range(1, minutes + 1):
            for m, label in skip_minutes:
                if i >= m and dp[i - m] + 1 < dp[i]:
                    dp[i] = dp[i - m] + 1
                    prev[i] = (i - m, label)
        skips = []
        cur = minutes
        while cur > 0 and prev[cur]:
            j, label = prev[cur]
            skips.append(label)
            cur = j
        used = sum(sec * skips.count(lbl) for sec, lbl in all_skip_values)
        rem = time - used
        if rem > 0:
            for sec, label in sorted(all_skip_values, key=lambda x: x[0]):
                if sec >= rem:
                    skips.append(label)
                    break
        
        return skips


from ..client.game_client import GameClient
from loguru import logger
import random





class Lords(GameClient):
    
    
    async def get_lords(self, sync: bool = True) -> dict | bool:
        
        
        try:
            
            await self.send_json_message("gli", {})
            
            if sync:
                response = await self.wait_for_response("gli")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False
        


    async def list_lords_id(
        self,
        lord_list: list,
    ) -> list:
        
        if not isinstance(lord_list, list):
            lord_list = []
        
        lords_data = await self.get_lords()
        all_lords = lords_data.get("C", [])
        
        for lord_obj in all_lords:
            lord_id = lord_obj.get("ID")
            eq_list = lord_obj.get("EQ")
            if len(eq_list) >= 5:
                lord_list.append(lord_id)


        
        
    async def select_lord(
        self,
        user_lords: list
    ) -> int:
        
        
        if not isinstance(user_lords, list):
            logger.error("Add lords list!")
            return 
        
        details_response = await self.send_rpc("gcl", {})
        account_id = details_response["PID"]
        moves_response = await self.send_rpc("gam", {})
        movements = [movement for movement in moves_response["M"] if movement["M"]["OID"] == account_id and movement.get("UM") is not None]
        used_lords = [movement["UM"]["L"].get("ID") for movement in movements]
        available_lords = list(set(user_lords) - set(used_lords))
        if available_lords:
            choosed_lord = random.choice(available_lords)
            return choosed_lord

        else:
            logger.warning("All lords are used!")
            return
        
        


from ..client.game_client import GameClient
from loguru import logger







class GlobalEffects(GameClient):
    
    
    
    async def get_global_effects(
        self,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("usg", {})
            if sync:
                response = await self.wait_for_response("usg")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def upgrade_global_effect(
        self,
        effect_id: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("agb", {"GEID": effect_id})
            if sync:
                response = await self.wait_for_response("agb")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False


from ..client.game_client import GameClient
from loguru import logger





class Tools(GameClient):
    
    
    async def get_production_queue(
        self, 
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("spl", {"LID": 1})
            if sync:
                response = await self.wait_for_response("spl")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            
            
            
    async def produce_tools(
        self,
        castle_id: int,
        wod_id: int,
        amount: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "bup",
                {
                    "LID": 1,
                    "WID": wod_id,
                    "AMT": amount,
                    "PO": -1,
                    "PWR": 0,
                    "SK": 73,
                    "SID": 0,
                    "AID": castle_id
                }
            )
            if sync:
                response = await self.wait_for_response("bup")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            
            
    async def cancel_production(
        self, 
        slot_type: str, 
        slot: int, 
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("mcu", {"LID": 1, "S": slot, "ST": slot_type})
            if sync:
                response = await self.wait_for_response("mcu")
                return response
            return True
        
        except Exception as e:
            logger.error(e)


import asyncio, time, random
from typing import Any, Dict, Optional
from dataclasses import dataclass
from loguru import logger







@dataclass
class _PlannedJob:
    key: str
    payload: Dict[str, Any]
    run_at_monotonic: float


class Planner:
    """
    Planner :
     - schedule(key, payload, interval, jitter) schedule serial task
     - start() run.
     - close() shutdown.
     - reset() reset all without worker
     - full_reset() reset full
    """
    def __init__(
        self,
        on_expire,
        *,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        process_interval: float = 2.0,
        max_proc: int = 2000
    ):
        
        if not asyncio.iscoroutinefunction(on_expire):
            raise TypeError("on_expire should be coroutine function")
        self.on_expire = on_expire
        self.loop = loop or asyncio.get_event_loop()
        self._queue: "asyncio.Queue[_PlannedJob]" = asyncio.Queue(maxsize=max_proc)
        self._last_sched: Dict[str, float] = {}      # key -> monotonic last scheduled run time
        self._timers: Dict[str, asyncio.Task] = {}   # key -> timer task (only last)
        self._worker_task: Optional[asyncio.Task] = None
        self._running = False
        self._shutdown = False
        self.process_interval = float(process_interval)


    @property
    def running(self) -> bool:
        return self._running and not self._shutdown

    @property
    def stopped(self) -> bool:
        return self._shutdown


    async def start(self):
        if self._running:
            return
        self._running = True
        self._worker_task = asyncio.create_task(self._worker_loop(), name="attack-planner-worker")


    async def close(self):
        
        self._shutdown = True
        for t in list(self._timers.values()):
            if not t.done():
                t.cancel()
        self._timers.clear()
        
        try:
            while True:
                self._queue.get_nowait()
                self._queue.task_done()
        
        except asyncio.QueueEmpty:
            pass
        
        
        if self._worker_task:
            await self._queue.put(_PlannedJob("__SENTINEL__", {}, time.monotonic()))
            try:
                await self._worker_task
            except Exception:
                pass
            self._worker_task = None
        
        self._last_sched.clear()


    async def reset(self):
        for t in list(self._timers.values()):
            if not t.done():
                t.cancel()
            
        self._timers.clear()
        try:
            while True:
                self._queue.get_nowait()
                self._queue.task_done()

        except asyncio.QueueEmpty:
            pass
        self._last_sched.clear()


    async def full_reset(self):
        await self.close()
        await self.start()


    async def schedule_after_last(self, key: str, payload: Dict[str, Any], *, interval: float = 1.0, jitter: float = 0.0):
 
        now = time.monotonic()
        last = self._last_sched.get(key, now)
        base = max(now, last)
        j = random.uniform(-jitter, jitter) if jitter and jitter > 0 else 0.0
        run_at = base + max(0.0, float(interval) + j)
        # update last scheduled time
        self._last_sched[key] = run_at

        # cancel previous timer for same key (we only care about last scheduled timer)
        prev = self._timers.get(key)
        if prev and not prev.done():
            prev.cancel()

        # schedule timer that will put planned job into the queue at run_at
        async def _timer():
            try:
                to_wait = run_at - time.monotonic()
                if to_wait > 0:
                    await asyncio.sleep(to_wait)
                if self._shutdown:
                    return
                await self._queue.put(_PlannedJob(key=key, payload=payload, run_at_monotonic=run_at))
            except asyncio.CancelledError:
                return
            except Exception:
                return
            finally:
                # remove timer reference when done
                self._timers.pop(key, None)

        t = asyncio.create_task(_timer(), name=f"attack-planner-timer-{key}")
        self._timers[key] = t
        return True



    async def _worker_loop(self):

        while True:
            job = await self._queue.get()
            if job.key == "__SENTINEL__":
                break
            try:
                await self.on_expire (job.key, job.payload)
            except asyncio.CancelledError:
                raise
            except Exception as e:
                # swallow to avoid killing worker; log externally if desired
                try:
                    
                    logger.exception("AttackPlanner: attack_cb raised: %s", e)
                except Exception:
                    pass
            # enforce interval between processing jobs
            if self.process_interval > 0:
                try:
                    await asyncio.sleep(self.process_interval)
                except asyncio.CancelledError:
                    raise
                except Exception:
                    pass
        return



from ..client.game_client import GameClient
from loguru import logger
from ..utils.utils import Utils
import asyncio






class Castle(GameClient):
    
    
    async def get_castles(self, sync: bool = True) -> dict | bool:
        
        try:
            
            await self.send_json_message("gcl", {})
            if sync:
                response = await self.wait_for_response("gcl")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False
        
        
    
      
    async def get_detailed_castles(self, sync: bool = True) -> dict | bool:
        
        try:
            
            await self.send_json_message("dcl", {})
            if sync:
                response = await self.wait_for_response("dcl")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False
        
        
    
       
    async def relocate_main_castle(
        self,
        x: int,
        y: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("rst", {"PX": x, "PY": y})
            if sync:
                response = await self.wait_for_response("rst")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False
                
    

    async def go_to_castle(
        self,
        kingdom: int,
        castle_id: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("jca", {"CID": castle_id, "KID": kingdom})
            if sync:
                response = await self.wait_for_response("jaa")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
    
    
    
    async def rename_castle(
        self,
        kingdom: int,
        castle_id: int,
        castle_type: int,
        name: str,
        paid: int = 0,
        sync: bool = True
    ) -> dict | bool:
        
        
        try:
            await self.send_json_message(
                "arc",
                {
                    "CID": castle_id,
                    "P": paid,
                    "KID": kingdom,
                    "AT": castle_type,
                    "N": name
                }
            )
            if sync:
                response = await self.wait_for_response("arc")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False
        
    
       
    async def get_castle_resources(self, sync: bool = True) -> dict | bool:
        
        try:
            
            await self.send_json_message("grc", {})
            if sync:
                response = await self.wait_for_response("grc")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False
    


    async def get_castle_production(self, sync: bool = True) -> dict | bool:
        
        try:
            
            await self.send_json_message("gpa", {})
            if sync:
                response = await self.wait_for_response("gpa")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False
    


    async def send_resources_to_kingdom(
        self,
        id_sender: int,
        sender_kid: int,
        target_kid: int,
        resources: list[list[str, int]],
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "kgt",
                {
                   "SCID": id_sender,
                   "SKID": sender_kid,
                   "TKID": target_kid,
                   "G": resources
                }
            )
            if sync:
                response = await self.wait_for_response("kgt")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
    async def send_units_to_kingdom(
        self,
        id_sender: int,
        sender_kid: int,
        target_kid: int,
        units: list[list[int, int]],
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "kut",
                {
                   "SCID": id_sender,
                   "SKID": sender_kid,
                   "TKID": target_kid,
                   "CID": -1,
                   "A": units
                }
            )
            if sync:
                response = await self.wait_for_response("kut")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def skip_kingdom_transfer(
        self,
        skip: str,
        target_kid: int,
        transfer_type: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "msk",
                {
                    "MST": skip,
                    "KID": target_kid,
                    "TT": transfer_type
                }
            )
            if sync:
                response = await self.wait_for_response("msk")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def auto_units_kingdom_transfer(
        self,
        id_sender: int,
        sender_kid: int,
        target_kid: int,
        units: list,
        skips: list = None,
        sync: bool = True  
    ) -> dict | bool:
        
        
        utils = Utils()
        send_units = await self.send_units_to_kingdom(id_sender, sender_kid, target_kid, units)
        if not isinstance(send_units, dict):
            logger.error(f"Failed to send units to kingdom: {target_kid}")
            return
        
        kpi = send_units.get("kpi", {})
        ut_list = kpi.get("UT")
        if not ut_list or not isinstance(ut_list, list):
            logger.error("Unknown transfer time data!")
            return
        
        time_to_transfer = ut_list[0].get("RS")
        if not isinstance(time_to_transfer, int):
            logger.error(f"Invalid time value: {ut_list[0]}")
            return
        
        skip_list = utils.skip_calculator(time_to_transfer, skips)
        for skip in skip_list:
            await self.skip_kingdom_transfer(skip, target_kid, transfer_type=1, sync=sync)
        
        logger.info(f"All units has been sent succesfully to kingdom {target_kid}!")
        
        
        
        
        
    async def auto_res_kingdom_transfer(
        self,
        id_sender: int,
        sender_kid: int,
        target_kid: int,
        resources: list,
        skips: list = None,
        sync: bool = True  
    ) -> dict | bool:
        
        
        utils = Utils()
        resources_sender = await self.send_resources_to_kingdom(id_sender, sender_kid, target_kid, resources)
        if not isinstance(resources_sender, dict):
            logger.error(f"Failed to send resources to kingdom: {target_kid}")
            return
        
        kpi = resources_sender.get("kpi", {})
        rt_list = kpi.get("RT")
        if not rt_list or not isinstance(rt_list, list):
            logger.error("Unknown transfer time data!")
            return
        
        time_to_transfer = rt_list[0].get("RS")
        if not isinstance(time_to_transfer, int):
            logger.error(f"Invalid time value: {rt_list[0]}")
            return

        skip_list = utils.skip_calculator(time_to_transfer, skips)
        for skip in skip_list:
            await self.skip_kingdom_transfer(skip, target_kid, transfer_type=2, sync=sync)
        
        logger.info(f"All resources has been sent succesfully to kingdom {target_kid}!")    





    async def units_replenish(
        self,
        target_kid: int,
        wod_id: int,
        amount: int
    ) -> None:
        
        
        try:
            
            account_inventory = await self.get_detailed_castles()
            inventory_data = account_inventory["C"]
            donors = []
            
            
            for kingdom in inventory_data:
                kid = kingdom.get("KID")
                if kid == target_kid:
                    continue
                
                for ai_block in kingdom.get("AI", []):
                    aid = ai_block.get("AID")
                    for wod, amt in ai_block.get("AC", []):
                        if wod == wod_id and amt > amount:
                            donors.append({"aid": aid, "kid":kid, "amount": amt})
                            break
                        
            if not donors:
                logger.warning("I can't find any eligible location!")
                return False
                            
            else:
                best = max(donors, key=lambda d: d["amount"])
                donor_aid = best["aid"]
                donor_amt = best["amount"]
                donor_kid = best["kid"]
                send_amt = min(donor_amt, amount)
                
                await self.auto_units_kingdom_transfer(donor_aid, donor_kid, target_kid, [[wod_id, send_amt]])
                logger.info(f"Kingdom {target_kid} refilled with {send_amt} units!")
                     
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def kingdom_auto_feeder(
        self,
        target_kid: int,
        min_food: int,
        min_mead: int,
        skips: list = None,
        interval: float = 60.0,
        max_transfers: int = 3,
        sync: bool = True,
        stop_event: asyncio.Event = None,
        min_donor_stock_food: int = 100_000,
        min_donor_stock_mead: int = 100_000,
    ) -> None:


        if stop_event is None:
            stop_event = asyncio.Event()

        while not stop_event.is_set():
            try:
                castles_inventory = await self.get_detailed_castles(sync=sync)
                resource_inventory = castles_inventory["C"]
                targets = [k for k in resource_inventory if k.get("KID") == target_kid]
                
                best_food_by_kid = {}
                best_mead_by_kid = {}
                
                for k in resource_inventory:
                    kid = k.get("KID")
                    if kid == target_kid:
                        continue
                    
                    for ai in k.get("AI", []):
                        aid = ai.get("AID")
                        gpa = ai.get("gpa", {}) or {}
                        food_val = int(ai.get("F", 0))
                        mead_val = int(ai.get("MEAD", 0))
                        food_prod = int(gpa.get("DF", 0))
                        mead_prod = int(gpa.get("DMEAD", 0))
                    
                        af = max(0, food_val - int(min_donor_stock_food)) if food_prod > 0 else 0
                        am = max(0, mead_val - int(min_donor_stock_mead)) if mead_prod > 0 else 0
                    
                        if af > 0:
                            cur = best_food_by_kid.get(kid)
                            if cur is None or af > cur["af"]:
                                best_food_by_kid[kid] = {"aid": aid, "af": af}
                            
                        if am > 0:
                            cur = best_mead_by_kid.get(kid)
                            if cur is None or am > cur["am"]:
                                best_mead_by_kid[kid] = {"aid": aid, "am": am}
                
                donors_food = sorted(
                    [(kid, d["aid"], d["af"]) for kid, d in best_food_by_kid.items()],
                    key=lambda t: t[2],
                    reverse=True
                )
                
                donors_mead = sorted(
                    [(kid, d["aid"], d["am"]) for kid, d in best_mead_by_kid.items()],
                    key=lambda t: t[2],
                    reverse=True
                )
                
                for t in targets:
                    kid = t.get("KID")
                    for ai in t.get("AI", []):
                        gpa = ai.get("gpa", {}) or {}
                        food_val = int(ai.get("F", 0))
                        mead_val = int(ai.get("MEAD", 0))
                        food_cap = int(gpa.get("MRF", 0))
                        mead_cap = int(gpa.get("MRMEAD", 0))
                        food_prod = int(gpa.get("DF", 0))
                        mead_prod = int(gpa.get("DMEAD", 0))
                        
                        food_deficit = max(0, (food_cap - 5) - food_val) if food_cap > 0 else 0
                        mead_deficit = max(0, (mead_cap - 5) - mead_val) if mead_cap > 0 else 0
                        need_food = (food_deficit > 0) and ((food_prod < 0) and (food_val < int(min_food)))
                        need_mead = (mead_deficit > 0) and ((mead_prod < 0) and (mead_val < int(min_mead)))
                        
                        # ——— FOOD ———
                        if need_food:
                            remaining = food_deficit
                            transfers_done = 0
                            for i, (dkid, daid, af) in enumerate(donors_food):
                                if remaining <= 0 or transfers_done >= max_transfers:
                                    break
                                if af <= 0:
                                    continue
                                send_amt = min(remaining, af)
                                try:
                                    await self.auto_res_kingdom_transfer(
                                        daid, dkid, kid, [["F", int(send_amt)]], skips, sync
                                    )
                                    remaining -= int(send_amt)
                                    transfers_done += 1
                                    donors_food[i] = (dkid, daid, af - int(send_amt))
                                except Exception as e:
                                    logger.error(f"[FOOD] transfer error {dkid}:{daid} -> {kid}: {e}")
                            logger.success(f"[FOOD] filled {food_deficit - remaining} / {food_deficit} towards KID {kid}")
                        
                        # ——— MEAD ———
                        if need_mead:
                            remaining = mead_deficit
                            transfers_done = 0
                            for i, (dkid, daid, am) in enumerate(donors_mead):
                                if remaining <= 0 or transfers_done >= max_transfers:
                                    break
                                if am <= 0:
                                    continue
                                send_amt = min(remaining, am)
                                try:
                                    await self.auto_res_kingdom_transfer(
                                        daid, dkid, kid, [["MEAD", int(send_amt)]], skips, sync
                                    )
                                    remaining -= int(send_amt)
                                    transfers_done += 1
                                    donors_mead[i] = (dkid, daid, am - int(send_amt))
                                except Exception as e:
                                    logger.error(f"[MEAD] transfer error {dkid}:{daid} -> {kid}: {e}")
                            
                            logger.success(f"[MEAD] filled {mead_deficit - remaining} / {mead_deficit} towards KID {kid}")
                            
            except asyncio.TimeoutError:
                logger.warning("Timeout occurred. Retry!")
            
            except Exception as e:
                logger.error(e)
            
            await asyncio.sleep(interval)
                          
                            
                


from ..client.game_client import GameClient
from loguru import logger






class Alliance(GameClient):
    
    
    
    async def get_chat(self, sync: bool = True) -> dict | bool:
        
        try:
            await self.send_json_message("acl", {})
            if sync:
                response = await self.wait_for_response("acl")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
    
    
    async def write_on_chat(
        self,
        message: str,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "acm",
                {
                    "M": message
                }
            )
            if sync:
                response = await self.wait_for_response("acm")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
    
    
    
       
        
    async def help_alliance_member(
        self, 
        kingdom: int, 
        help_id: int, 
        sync: bool = True
    ) -> dict | bool:

        try:
            await self.send_json_message("ahc", {"LID": help_id, "KID": kingdom})
            if sync:
                response = await self.wait_for_response("ahc")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
    
    
    
       
    async def help_alliance_all(
        self, 
        kingdom: int, 
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("aha", {"KID": kingdom})
            if sync:
                response = await self.wait_for_response("aha")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False  
        
        
    async def invite_player(
        self,
        user_id: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "aip",
                {
                    "SV": user_id
                }
            )
            if sync:
                response = await self.wait_for_response("aip")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False 
        
        
        
    async def rank_player(
        self,
        account_id: int,
        rank: int,   ## 0-8 leader to member
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "arm",
                {
                    "PID": account_id,
                    "R": rank
                }  
            )
            if sync:
                response = await self.wait_for_response("arm")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False 
        
      
      
        
        
    async def mass_message(
        self,
        text: str,
        title: str = None,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "anl",
                {
                    "SJ": title,
                    "TXT": text
                }
            )
            if sync:
                response = await self.wait_for_response("anl")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False 
        
        
        
    async def leave_alliance(self, sync: bool = True) -> dict | bool:
        
        try:
            await self.send_json_message("aqi", {})
            if sync:
                response = await self.wait_for_response("aqi")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False 
        
        
        
    async def create_alliance(
        self,
        alliance_name: str,
        allinace_state: int,
        description: str,
        lang: str,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "afo",
                {
                    "PO":-1,
                    "PWR":0,
                    "IA": allinace_state,
                    "N": alliance_name,
                    "D": description,
                    "ALL": lang
                }
            )
            if sync:
                response = await self.wait_for_response("afo")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False 
        
        
    async def alliance_info(
        self,
        alliance_id: int,
        sync: bool = True
        ) -> dict | bool:
        
        try:
            await self.send_json_message("ain", {"AID": alliance_id})
            if sync:
                response = await self.wait_for_response("ain")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False 
        
        
    async def get_alliance_discussions_panel(self, sync: bool = True) -> dict | bool:
        
        try:
            await self.send_json_message("gat", {})
            if sync:
                response = await self.wait_for_response("gat")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False 
        
        
    async def post_alliance_announce(
        self,
        announce_text: str, 
        sync: bool = True
        ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "acd",
                {
                    "T": 0,
                    "TXT": announce_text
                }
            )
            if sync:
                response = await self.wait_for_response("gat")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False 
        
       
        
    async def postin_alliance_discussions_panel(
        self,
        title: str,
        text: str,
        rank_list: list = [0, 1, 2],
        sync: bool = True
        ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "atc",
                {
                    "N": title,
                    "RG": rank_list,
                    "R": text
                }
            )
            if sync:
                response = await self.wait_for_response("gat")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False 


from ..client.game_client import GameClient
from loguru import logger





class Account(GameClient):





    async def get_account_infos(self, sync: bool = True) -> dict | bool:

        try:
            await self.send_json_message("gpi", {})
            if sync:
                response = await self.wait_for_response("gpi")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False
        



    async def register_email(
        self,
        email: str,
        subscribe: bool = False,
        sync: bool = True
    ) -> dict | bool:

        try:
            await self.send_json_message("vpm", {"MAIL": email, "NEWSLETTER": subscribe})
            if sync:
                response = await self.wait_for_response("vpm")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False




    async def get_username_change_infos(
        self, sync: bool = True
    ) -> dict | bool:

        try:
            await self.send_json_message("gnci", {})
            if sync:
                response = await self.wait_for_response("gnci")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False




    async def change_username(
        self, new_username: str, sync: bool = True
    ) -> dict | bool:

        try:
            await self.send_json_message("cpne", {"PN": new_username})
            if sync:
                response = await self.wait_for_response("cpne")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False




    async def change_password(
        self,
        old_password: str,
        new_password: str,
        sync: bool = True
    ) -> dict | bool:

        try:
            await self.send_json_message("scp", {"OPW": old_password, "NPW": new_password})
            if sync:
                response = await self.wait_for_response("scp")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False




    async def ask_email_change(
        self, new_email: str, sync: bool = True
    ) -> dict | bool:

        try:
            await self.send_json_message("rmc", {"PMA": new_email})
            if sync:
                response = await self.wait_for_response("rmc")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False



    async def get_email_change_status(
        self, sync: bool = True
    ) -> dict | bool:

        try:
            await self.send_json_message("mns", {})
            if sync:
                response = await self.wait_for_response("mns")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False



    async def cancel_email_change(
        self, sync: bool = True
    ) -> dict | bool:

        try:
            await self.send_json_message("cmc", {})
            if sync:
                response = await self.wait_for_response("cmc")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False